//-----------------------------------------------------------------------------
// File:    WristFlexionExtensionMuscleModelRigidBody.cpp
// Class:    None
// Parent:   None
// Children: None
// Purpose:  Simulates a hand (rigid body) moving with respect to the forearm (ground).  The hand rotates about one
//           principal axis, the body-fixed flexion-extension axis.
//-----------------------------------------------------------------------------
// The following are standard C/C++ header files.
// If a filename is enclosed inside < >  it means the header file is in the Include directory.
// If a filename is enclosed inside " "  it means the header file is in the current directory.
#include <ctype.h>      // Character Types
#include <math.h>       // Mathematical Constants
#include <stdarg.h>     // Variable Argument Lists
#include <stdio.h>      // Standard Input/Output Functions
#include <stdlib.h>     // Utility Functions
#include <string.h>     // String Operations
#include <signal.h>     // Signals (Contol-C + Unix System Calls)
#include <setjmp.h>     // Nonlocal Goto (For Control-C)
#include <time.h>       // Time and Date information
#include <assert.h>     // Verify Program Assertion
#include <errno.h>      // Error Codes (Used in Unix system())
#include <float.h>      // Floating Point Constants
#include <limits.h>     // Implementation Constants
#include <stddef.h>     // Standard Definitions
#include <exception>    // Exception handling (e.g., try, catch throw)
#include "SimTKsimbody.h"
#include "ScaledBodySegmentMassProperties.h"
using namespace SimTK;
using namespace std;
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Prototypes for local functions (functions not called by code in other files)
//-----------------------------------------------------------------------------
bool  SimulateWristFlexionExtension( void );
bool  WriteStringToFile(   const char outputString[], FILE *fptr )  { return fputs( outputString, fptr ) != 0; }
bool  WriteStringToScreen( const char outputString[] )              { return WriteStringToFile( outputString, stdout ); }
bool  WriteDoubleToFile( double x, int precision, FILE *fptr );
FILE*  FileOpenWithMessageIfCannotOpen( const char *filename, const char *attribute );


//-----------------------------------------------------------------------------
// The executable program starts here
//-----------------------------------------------------------------------------
int  main( int numberOfCommandLineArguments, char *arrayOfCommandLineArguments[] )
{
   // Default value is simulation failed
   bool simulationSucceeded = false;

   try
   {
      // Simulate the multibody system
     simulationSucceeded = SimulateWristFlexionExtension();
   }
   
   // This catch statement handles certain types of expections
   catch( const exception &e )
   {
      WriteStringToScreen("\n\nError: Programming error encountered. \n The exception thrown is: " );
	   WriteStringToScreen( e.what() );
   }

   // The exception-declaration statement (...) handles any types of excpetion,
   // including C exceptions an system/application generated exception.
   // This includes exceptions such as memory protection and floating-point violations.
   // An ellipsis catch handler must be the last handler for its try block.
   catch( ... )
   {  
      WriteStringToScreen( "\n\n Errorr:  Programming error encountered. \n An  unhandled expection was thrown." );
   }

   // Give the user a chance to see a final program message before exiting the program
   WriteStringToScreen(simulationSucceeded ? "\n\n Simulatin succeeded.  Press Enter to terminate the program: " : "\n\nSimulation failed. Press Enter to terminate the program: " );
   
   // Keep the screen displayed until the user presses the Enter (or Return) key.
   getchar();

   // The value returned by the main function is the exit status of the program.
   // A normal program exit returns 0 (other return values usually signal an error).
   return simulationSucceeded == true ? 0 : 1;
}



//-----------------------------------------------------------------------------
bool  WristFlexionExtension( void )
{
   // Declare a multibody system (contains one or more force and matter sub-systems)
   MultibodySystem  mbs;

   // Create a matter sub-system (the hand)
   SimbodyMatterSubsystem  hand;

   // Create the mass, center of mass, and inertia properties for the hand
   
   // Enter the subject-specific total mass and total height
   const Real subjectTotalMassInKG = 90.0;
   const Real subjectTotalHeightInMM = 0.0;
   const bool subjectIsFemale = false;


   // Note: A value of zero for a segment length causes the program to estimate the segment length using the
   //       standardBodySegmentLengthInMM * subjectTotalHeightInMM / standardBodyTotalHeightInMM.
   const char *segmentName[] =      { "Hand" };
   const Real segmentLengthInMM[] = { 180.4 };

  // Get the standard body segment mass properties
   const StandardBodySegmentMassProperties*  standardSegment =  StandardBodySegmentMassProperties::getStandardHumanBodySegmentMassProperties(subjectIsFemale, segmentName);
   // Look through the list for the hand symbol
   if( !standardSegment ) return false;

   //  This is a suitcase for the subject specific Scaled Mass Properties

   const ScaledBodySegmentMassProperties scaledSegment( *standardSegment,subjectTotalMassInKG, subjectTotalHeightInMM, subjectSegmentLengthInMM );

   // Scale the properties with the given information:
   // Now with the get commands, retrieve the scaled properties 
   Real segmentLength = scaledSegment.getSegmentLengthInMM();

   // Write the segment mass to the file
   Real segmentMass = scaledSegment.getMassInKG();

   // Get the segment's distance from its origin anatomical marker to its mass center distance
   Real segmentMassCenterLength = scaledSegment.getMassCenterLengthInMM();

   // Get       the segment x-radius of gyration about its mass center to the file
   // Calculate the segment x-moment of inertia  about its mass center to the file
   Real kxx = scaledSegment.getXRadiusOfGyrationInMM();
   Real Ixx = kxx*kxx*segmentMass;

   // Get       the segment y-radius of gyration about its mass center to the file
   // Calculate the segment y-moment of inertia  about its mass center to the file
   Real kyy = scaledSegment.getYRadiusOfGyrationInMM();
   Real Iyy = kyy*kyy*segmentMass;

   // Get       the segment z-radius of gyration about its mass center to the file
   // Calculate the segment z-moment of inertia  about its mass center to the file
   Real kzz = scaledSegment.getZRadiusOfGyrationInMM();
   Real Izz = kzz*kzz*segmentMass;

   // The location of the hand's center of mass is a vector from the hand's
   // origin expressed in the "x, y, z"' unit vectors fixed in the hand's frame.
   // Example: The vector(0,0,0) locates the hand's center of mass at the hand's origin.
   // Example: The vector(1,0,0) locates the hand's center of mass 1 unit in the "x" direction from the hand's origin.
   const Vec3  handCenterOfMassLocation( 0, 0, 0 );

   // Create the hand's inertia matrix about its origin for the "x, y, z" unit vectors fixed in the hand's frame.
   // Note: The 3x3 inertia matrix is symmetric - so only 6 elements need to be defined.
   // Ixx, Iyy, Izz  are moments  of inertia (    diagonal terms in the matrix)
   // Ixy, Ixz, Iyz  are products of inertia (off-diagonal terms in the matrix)
   const Real  Ixy = 0,  Ixz = 0,   Iyz = 0;
   const Inertia  handInertiaMatrix( Ixx, Iyy, Izz, Ixy, Ixz, Iyz );

   // The MassProperties class holds the mass, center of mass, and inertia properties of a rigid body.
   // Although the next line creates an instance of the MassProperties class for the hand,
   // it does not get associated with the hand until the  addRigidBody  method.
   MassProperties  handMassProperties( segmentMass, handCenterOfMassLocation, handInertiaMatrix );

   // The motion of the hand is related to the motion of the ground via "mobilizers"
   // The "mobilizers" specify the allowable motion of the hand to the ground.
   // More specifically, the motion of an "outboard body" (e.g., the hand)
   // to its inboard body (e.g., the ground) is specified by first constructing:
   // 1. An  "outboard frame"  on the ground (which hooks to the "outboard body" - the hand )
   // 2. An  "inboard  frame"  on the hand  (which hooks to the "inboard  body" - the ground)

   // The orientation and position of the  outboard frame  from the ground's frame is specified below.
   // The outboard frame's axes are aligned with the ground's axes and its origin is coincident with the ground's origin.
   const Transform outboardFrameTransformFromGround;  // The default constructor is the identity transform

   // The orientation and position of the  inboard  frame  from the hand's  frame is specified below.
   // The inboard frame's  axes are aligned with the hand's  axes and its origin is coincident with the hand's  origin
   // In other words, for this simple problem the inboard  frame and the hand frame are identical.
   // Although the inboard frame can be constructed in a simple manner analogous to the outboardFrameTransformFromGround (above)
   // it is worthwhile to look at the details of the rotation matrix and position vector in the transform:
   // a. The rotation matrix relating the InboardFrame's x,y,z axes to the handFrame's x,y,z axes is specified InboardFrame_handFrame.
   // b. The position of the InboardFrame's origin from the handFrame origin, expressed in terms of the handFrame's "x, y, z" unit vectors.
   const Rotation   inboardFrameOrientationInHand;   // ( 1,0,0, 0,1,0, 0,0,1 );
   const Vec3       inboardFrameOriginLocationFromHandOrigin( 0, segmentMassCenterLength, 0 );
   const Transform  inboardFrameTransformFromHand( inboardFrameOrientationInHand, inboardFrameOriginLocationFromHandOrigin );

   // There are many ways that the hand can move relative to the ground.
   // The following allows the hand to move in "x", "y", and "z" directions.
   // Another option producing the same result is Mobilizer::Free
   Mobilizer handToGroundMobilizer = Mobilizer::Gimbal;
   const BodyId handBodyId = hand.addRigidBody( handMassProperties, inboardFrameTransformFromHand, GroundId, outboardFrameTransformFromGround, handToGroundMobilizer );

   // Add the matter (hand) sub-system to the system.
   mbs.setMatterSubsystem( hand );

  //TO DO......
   return true;
}


//-----------------------------------------------------------------------------
FILE*  FileOpenWithMessageIfCannotOpen( const char *filename, const char *attribute )
{
   // Try to open the file
   FILE *Fptr1 = fopen( filename, attribute );

   // If unable to open the file, issue a message
   if( !Fptr1 )
   {
      WriteStringToScreen( "\n\n Unable to open the file: " );
      WriteStringToScreen( filename );
      WriteStringToScreen( "\n\n" );
   }

   return Fptr1;
}


//-----------------------------------------------------------------------------
bool  WriteDoubleToFile( double x, int precision, FILE *fptr )
{
   // Ensure the precision (number of digits in the mantissa after the decimal point) makes sense.
   // Next, calculate the field width so it includes one extra space to the right of the number.
   if( precision < 0 || precision > 17 ) precision = 5;
   int fieldWidth = precision + 8;

   // Create the format specifier and print the number
   char format[20];
   sprintf( format, " %%- %d.%dE", fieldWidth, precision );
   return fprintf( fptr, format, x ) >= 0;
}

